import os

_CURRENT_DIR = os.path.dirname(__file__)


class Bitmaps:
    SMILEY = os.path.join(_CURRENT_DIR, '../bitmaps/smiley.bitmap.json')  # noqa


class Movies:
    PING_PONG = os.path.join(_CURRENT_DIR, '../movies/ping_pong.movie.json')  # noqa
